/***************************************************************************
                           csearchindex.cpp  -  description
                             -------------------
    begin                : Mon May 14 2003
    copyright            : (C) 2003-2004 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include <string.h>

#include "dcos.h"
#include "core/cbytearray.h"
#include "cconfig.h"
#include "core/cstring.h"
#include "core/cdir.h"
#include "core/cfile.h"
#include "cdownloadmanager.h"
#include "dclib.h"
#include "core/cmanager.h"
#include "cfilemanager.h"
#include "core/cbase32.h"

#include "hash/compat.h"
#include "hash/MerkleTree.h"

#include "cfilehasher.h"

#include "csearchindex.h"

/** */
CSearchIndex::CSearchIndex()
{
	m_pBaseArray       = new CByteArray();
	m_pFileBaseArray   = new CByteArray();
	m_pPathBaseArray   = new CByteArray();
	
	m_pHashBaseArray = new CByteArray();
	m_pHashFileBaseArray = new CByteArray();
	m_pHashPathBaseArray = new CByteArray();
	m_pHashIndex = new CByteArray();
	
	m_pCaseFoldedBase = new CByteArray();
	m_pCaseFoldedData = new CByteArray();
	
	m_pNewHashLeaves = 0;
	hashleavessize = 0; // this will get set when m_pNewHashLeaves is created
	
	/* during list refresh, these are created and updated
	 * and replace the old byte arrays when refresh stage is finished
	 */
	m_pUpdatingFileBaseArray     = 0;
	m_pUpdatingPathBaseArray     = 0;
	m_pUpdatingBaseArray         = 0;
	
	m_pUpdatingHashBaseArray     = 0;
	m_pUpdatingHashFileBaseArray = 0;
	m_pUpdatingHashPathBaseArray = 0;
	m_pUpdatingHashIndex         = 0;

	m_pUpdatingCaseFoldedBase    = 0;
	m_pUpdatingCaseFoldedData    = 0;

	m_bLoadedOK = LoadIndex();
}

/** */
CSearchIndex::~CSearchIndex()
{
	delete m_pBaseArray;
	delete m_pFileBaseArray;
	delete m_pPathBaseArray;
	
	delete m_pHashBaseArray;
	delete m_pHashFileBaseArray;
	delete m_pHashPathBaseArray;
	delete m_pHashIndex;
	
	delete m_pCaseFoldedBase;
	delete m_pCaseFoldedData;
	
	if (m_pNewHashLeaves != 0)
	{
		DPRINTF("~CSearchIndex: warning new hash leaves not saved!\n");
		SaveNewHashLeaves();
		
		if ( m_pNewHashLeaves != 0 ) // something went wrong in SaveNewHashLeaves()
		{
			delete m_pNewHashLeaves;
		}
	}
	
	if ( m_pUpdatingFileBaseArray ) delete m_pUpdatingFileBaseArray;
	if ( m_pUpdatingPathBaseArray ) delete m_pUpdatingPathBaseArray;
	if ( m_pUpdatingBaseArray ) delete m_pUpdatingBaseArray;
	
	if ( m_pUpdatingHashBaseArray ) delete m_pUpdatingHashBaseArray;
	if ( m_pUpdatingHashFileBaseArray ) delete m_pUpdatingHashFileBaseArray;
	if ( m_pUpdatingHashPathBaseArray ) delete m_pUpdatingHashPathBaseArray;
	if ( m_pUpdatingHashIndex ) delete m_pUpdatingHashIndex;
	
	if ( m_pUpdatingCaseFoldedBase ) delete m_pUpdatingCaseFoldedBase;
	if ( m_pUpdatingCaseFoldedData ) delete m_pUpdatingCaseFoldedData;
}

/** */
void CSearchIndex::PrepareUpdate()
{
	// files list
	if ( m_pUpdatingFileBaseArray ) delete m_pUpdatingFileBaseArray;
	if ( m_pUpdatingPathBaseArray ) delete m_pUpdatingPathBaseArray;
	if ( m_pUpdatingBaseArray ) delete m_pUpdatingBaseArray;
	
	m_pUpdatingFileBaseArray = new CByteArray(0);
	m_pUpdatingPathBaseArray = new CByteArray(0);
	m_pUpdatingBaseArray     = new CByteArray(0);
	
	// search lists of case folded path+filenames
	if ( m_pUpdatingCaseFoldedBase ) delete m_pUpdatingCaseFoldedBase;
	if ( m_pUpdatingCaseFoldedData ) delete m_pUpdatingCaseFoldedData;
	
	m_pUpdatingCaseFoldedBase = new CByteArray(0);
	m_pUpdatingCaseFoldedData = new CByteArray(0);
	
	// hash lists
	if ( m_pUpdatingHashBaseArray ) delete m_pUpdatingHashBaseArray;
	if ( m_pUpdatingHashFileBaseArray ) delete m_pUpdatingHashFileBaseArray;
	if ( m_pUpdatingHashPathBaseArray ) delete m_pUpdatingHashPathBaseArray;
	if ( m_pUpdatingHashIndex ) delete m_pUpdatingHashIndex;
	
	m_pUpdatingHashBaseArray = new CByteArray(0);
	m_pUpdatingHashFileBaseArray = new CByteArray(0);
	m_pUpdatingHashPathBaseArray = new CByteArray(0);
	m_pUpdatingHashIndex = new CByteArray(0);
	
	m_pUpdatingHashBaseArray->Append( m_pHashBaseArray->Data(), m_pHashBaseArray->Size() );
	m_pUpdatingHashFileBaseArray->Append( m_pHashFileBaseArray->Data(), m_pHashFileBaseArray->Size() );
	m_pUpdatingHashPathBaseArray->Append( m_pHashPathBaseArray->Data(), m_pHashPathBaseArray->Size() );
	m_pUpdatingHashIndex->Append( m_pHashIndex->Data(), m_pHashIndex->Size() );
}

/** */
void CSearchIndex::FinishUpdate()
{
	CByteArray * old1  = m_pFileBaseArray;
	CByteArray * old2  = m_pPathBaseArray;
	CByteArray * old3  = m_pBaseArray;
	CByteArray * old4  = m_pCaseFoldedBase;
	CByteArray * old5  = m_pCaseFoldedData;
	CByteArray * old7  = m_pHashBaseArray;
	CByteArray * old8  = m_pHashFileBaseArray;
	CByteArray * old9  = m_pHashPathBaseArray;
	CByteArray * old10 = m_pHashIndex;
	
	// swap to new lists / append new data - this section needs to be as fast as possible
	// essentially all operations need to be done simultaneously
	// it used to have some data appending, but now
	// it's just 7 pointer swaps, so it should be OK
	m_pFileBaseArray     = m_pUpdatingFileBaseArray;
	m_pPathBaseArray     = m_pUpdatingPathBaseArray;
	m_pBaseArray         = m_pUpdatingBaseArray;
	m_pCaseFoldedBase    = m_pUpdatingCaseFoldedBase;
	m_pCaseFoldedData    = m_pUpdatingCaseFoldedData;
	m_pHashFileBaseArray = m_pUpdatingHashFileBaseArray;
	m_pHashPathBaseArray = m_pUpdatingHashPathBaseArray;
	m_pHashIndex         = m_pUpdatingHashIndex;
	m_pHashBaseArray     = m_pUpdatingHashBaseArray;
	// end of speed critical section
	
	m_pUpdatingFileBaseArray     = 0;
	m_pUpdatingPathBaseArray     = 0;
	m_pUpdatingBaseArray         = 0;
	m_pUpdatingCaseFoldedBase    = 0;
	m_pUpdatingCaseFoldedData    = 0;
	m_pUpdatingHashBaseArray     = 0;
	m_pUpdatingHashFileBaseArray = 0;
	m_pUpdatingHashPathBaseArray = 0;
	m_pUpdatingHashIndex         = 0;

	delete old1;
	delete old2;
	delete old3;
	delete old4;
	delete old5;
	delete old7;
	delete old8;
	delete old9;
	delete old10;
}

/** */
void CSearchIndex::ResetHashIndex()
{
	m_pHashBaseArray->SetSize(0);
	m_pHashFileBaseArray->SetSize(0);
	m_pHashPathBaseArray->SetSize(0);
	m_pHashIndex->SetSize(0);
	
	if ( m_pNewHashLeaves == 0 )
	{
		m_pNewHashLeaves = new CByteArray();
	}
	m_pNewHashLeaves->SetSize(0);
	m_pNewHashLeaves->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashleaves.bin");
	delete m_pNewHashLeaves;
	m_pNewHashLeaves = 0;
}

/** */
std::set<unsigned long> * CSearchIndex::SearchHash( unsigned char * hash )
{
	unsigned long hbi,hi,bi;
	std::set<unsigned long> * results = 0;

	hi = 0;
	
	while ( FindHash( hash, &hi ) )
	{
		if ( HashBaseIndexFromHashIndex( hi, &hbi ) )
		{
			if ( BaseIndexFromHashBaseIndex( hbi, &bi ) )
			{
				if ( !results )
				{
					results = new std::set<unsigned long>;
				}
				results->insert(bi);
			}
		}
		hi+=dcpp::TigerTree::BYTES;
	}

	return results;
}

/** */
bool CSearchIndex::GetFileBaseObject( unsigned long index, struct filebaseobject * fbo )
{
	bool res = false;

	if ( (index*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
	{
		memcpy( fbo, m_pBaseArray->Data()+(index*sizeof(struct filebaseobject)), sizeof(struct filebaseobject) );

		res = true;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObjectDuringUpdate( unsigned long index, struct filebaseobject * fbo )
{
	bool res = false;

	if ( (index*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		memcpy( fbo, m_pUpdatingBaseArray->Data()+(index*sizeof(struct filebaseobject)), sizeof(struct filebaseobject) );

		res = true;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObject( unsigned long index, struct filebaseobject * fbo, CString & filename )
{
	bool res = false;

	if ( GetFileBaseObject(index,fbo) )
	{
		if ( m_pPathBaseArray->Size() > fbo->m_nPathIndex )
		{
			filename = (char*)m_pPathBaseArray->Data()+fbo->m_nPathIndex;

			if ( filename.NotEmpty() )
			{
				filename += DIRSEPARATOR;
			}
		}

		if ( m_pFileBaseArray->Size() > fbo->m_nFileIndex )
		{
			filename += (char*)m_pFileBaseArray->Data()+fbo->m_nFileIndex;
		}

		res = true;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObjectDuringUpdate( unsigned long index, struct filebaseobject * fbo, CString & filename )
{
	bool res = false;

	if ( GetFileBaseObjectDuringUpdate(index,fbo) )
	{
		if ( m_pUpdatingPathBaseArray->Size() > fbo->m_nPathIndex )
		{
			filename = (char*)m_pUpdatingPathBaseArray->Data()+fbo->m_nPathIndex;

			if ( filename.NotEmpty() )
			{
				filename += DIRSEPARATOR;
			}
		}

		if ( m_pUpdatingFileBaseArray->Size() > fbo->m_nFileIndex )
		{
			filename += (char*)m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex;
		}

		res = true;
	}

	return res;
}

/** */
bool CSearchIndex::GetCaseFoldedName( unsigned long index, struct filebaseobject * fbo, CString & name )
{
	unsigned long bapos = index*sizeof(struct filebaseobject);
	unsigned long cfpos = index*sizeof(unsigned long);
	
	if ( (bapos < m_pBaseArray->Size()) && (cfpos < m_pCaseFoldedBase->Size()) )
	{
		memcpy( fbo, m_pBaseArray->Data()+bapos, sizeof(struct filebaseobject) );
		
		unsigned long cfi;
		
		memcpy( &cfi, m_pCaseFoldedBase->Data()+cfpos, sizeof(unsigned long) );
		
		if ( cfi < m_pCaseFoldedData->Size() )
		{
			name = (char*) m_pCaseFoldedData->Data()+cfi;
			return true;
		}
	}
	
	return false;
}

/** */
bool CSearchIndex::LoadIndex()
{
	bool err = false;
	CDir d;
	ulonglong filesize;

	// first try to load hash index
	if ( m_pHashBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashbase.bin") == false )
	{
		err = true;
	}
	if ( err == false )
	{
		if ( m_pHashFileBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashfilebase.bin") == false )
		{
			err = true;
		}
	}
	if ( err == false )
	{
		if ( m_pHashPathBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashpathbase.bin") == false )
		{
			err = true;
		}
	}
	if ( err == false )
	{
		if ( m_pHashIndex->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashindex.bin") == false )
		{
			err = true;
		}
	}
	/* if ( err == false )
	{
		if ( OpenHashLeaves() == false )
		{
			err = true;
		}
		else
		{
			CloseHashLeaves();
		} */
		
		/* Just check hashleaves.bin is present and not zero bytes */
		/* Don't even bother doing that anymore, it may be zero bytes if nothing shared or hashing disabled */
		/* if ( (CConfig::Instance()->GetDisableHashList() == false) &&
		     (d.getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",false) <= 0)
		   )
		{
			err = true;
		}
	} */
	
	if ( err == true )
	{
		ResetHashIndex();
		
		err = false;
	}
	
	// sanity check
	filesize = d.getFileSize(CConfig::Instance()->GetConfigPath()+"database.bin",false);

	if ( (filesize%sizeof(struct filebaseobject)) != 0 )
	{
		err = true;
	}

	// load the filelist
	if ( err == false )
	{
		if ( m_pBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"database.bin") == false )
		{
			err = true;
		}
	}

	if ( err == false )
	{
		if ( m_pFileBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"filebase.bin") == false )
		{
			err = true;
		}
	}

	if ( err == false )
	{
		if ( m_pPathBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"pathbase.bin") == false )
		{
			err = true;
		}
	}

	if ( err )
	{
		m_pBaseArray->SetSize(0);
		m_pFileBaseArray->SetSize(0);
		m_pPathBaseArray->SetSize(0);
	}
	
	if ( !err )
	{
		filesize = d.getFileSize(CConfig::Instance()->GetConfigPath()+"casefoldedbase.bin");
		if ( filesize%sizeof(unsigned long) != 0 )
		{
			err = true;
		}
		
		if ( !err )
		{
			err = !(m_pCaseFoldedBase->LoadFromFile(CConfig::Instance()->GetConfigPath()+"casefoldedbase.bin"));
		}
		
		if ( !err )
		{
			err = !(m_pCaseFoldedData->LoadFromFile(CConfig::Instance()->GetConfigPath()+"casefoldeddata.bin"));
		}
	}

	return !err;
}

/** */
void CSearchIndex::SaveIndex()
{
	m_pBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"database.bin");
	m_pFileBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"filebase.bin");
	m_pPathBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"pathbase.bin");
	
	m_pCaseFoldedBase->SaveToFile(CConfig::Instance()->GetConfigPath()+"casefoldedbase.bin");
	m_pCaseFoldedData->SaveToFile(CConfig::Instance()->GetConfigPath()+"casefoldeddata.bin");
	
	m_pHashBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashbase.bin");
	m_pHashFileBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashfilebase.bin");
	m_pHashPathBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashpathbase.bin");
	m_pHashIndex->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashindex.bin");
	
	if (m_pNewHashLeaves != 0)
	{
		SaveNewHashLeaves();
	}
}

/** */
unsigned long CSearchIndex::IndexCount()
{
	unsigned long i = 0;

	if ( m_pBaseArray )
	{
		i = m_pBaseArray->Size()/sizeof(struct filebaseobject);
	}

	return i;
}

/** */
unsigned long CSearchIndex::IndexCountDuringUpdate()
{
	unsigned long i = 0;

	if ( m_pUpdatingBaseArray )
	{
		i = m_pUpdatingBaseArray->Size()/sizeof(struct filebaseobject);
	}

	return i;
}

/** */
unsigned long CSearchIndex::AddIndex( CFileInfo *fileinfo, CString path, eFileTypes filetype )
{
	struct filebaseobject fbo;
	unsigned long i;

	fbo.m_eFileType  = filetype;
	fbo.m_nFileIndex = m_pUpdatingFileBaseArray->Size();
	fbo.m_nPathIndex = m_pUpdatingPathBaseArray->Size();
	fbo.m_nHashIndex = (unsigned long)-1;
	fbo.m_nSize      = fileinfo->size;
	fbo.m_tModTime   = fileinfo->st_m_time;

	i = m_pUpdatingBaseArray->Size()/sizeof(struct filebaseobject);

	m_pUpdatingBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
	m_pUpdatingFileBaseArray->Append( fileinfo->name.Data(), fileinfo->name.Length()+1 );
	m_pUpdatingPathBaseArray->Append( path.Data(), path.Length()+1 );
	
	return i;
}

/** Only used during update */
void CSearchIndex::AddSearchIndex( const CString & data )
{
	unsigned long pos = m_pUpdatingCaseFoldedData->Size();
	
	m_pUpdatingCaseFoldedBase->Append( (const char*)&pos, sizeof(unsigned long) );
	m_pUpdatingCaseFoldedData->Append( data.Data(), data.Length() + 1 );
}

/** UpdateIndex is only used during update */
void CSearchIndex::UpdateIndex( unsigned long index, struct filebaseobject * fbo )
{
	if ( (index*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		memcpy( m_pUpdatingBaseArray->Data()+(index*sizeof(struct filebaseobject)), fbo, sizeof(struct filebaseobject) );
	}
}

/** */
bool CSearchIndex::FindHash( unsigned char * hash, unsigned long * hi )
{
	unsigned long i;
	
	for(i=*hi;i<m_pHashIndex->Size();i+=dcpp::TigerTree::BYTES)
		if ( memcmp(hash,m_pHashIndex->Data()+i,dcpp::TigerTree::BYTES) == 0 )
		{
			*hi = i;
			return true;
		}
		
	return false;
}

/** */
bool CSearchIndex::FindHashDuringUpdate( unsigned char * hash, unsigned long * hi )
{
	unsigned long i;
	
	for(i=*hi;i<m_pUpdatingHashIndex->Size();i+=dcpp::TigerTree::BYTES)
		if ( memcmp(hash,m_pUpdatingHashIndex->Data()+i,dcpp::TigerTree::BYTES) == 0 )
		{
			*hi = i;
			return true;
		}
		
	return false;
}

/** */
CString CSearchIndex::GetHash( unsigned long hbi )
{
	CString s;
	CBase32 base32;
	CByteArray dst,src;
	struct hashbaseobject * hbo;

	if ( hbi < m_pHashBaseArray->Size() )
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+hbi);
		src.Append(m_pHashIndex->Data()+hbo->m_nHashIndex,dcpp::TigerTree::BYTES);
		base32.Encode( &dst, &src );
		if ( dst.Size() > 0 )
		{
			s.Set((const char*)dst.Data(),dst.Size());
		}
	}
	
	return s;
}

/** */
bool CSearchIndex::HashBaseIndexFromHashIndex( unsigned long hi, unsigned long * hbi )
{
	unsigned long i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+i);
		if ( hbo->m_nHashIndex == hi )
		{
			*hbi = i;
			return true;
		}
	}
			
	return false;
}

/** */
bool CSearchIndex::HashBaseIndexFromHashIndexDuringUpdate( unsigned long hi, unsigned long * hbi )
{
	unsigned long i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pUpdatingHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+i);
		if ( hbo->m_nHashIndex == hi )
		{
			*hbi = i;
			return true;
		}
	}
			
	return false;
}

/** */
bool CSearchIndex::BaseIndexFromHashBaseIndex( unsigned long hbi, unsigned long * bi )
{
	unsigned long i;
	struct filebaseobject * fbo;
	
	for(i=0;i<m_pBaseArray->Size();i+=sizeof(struct filebaseobject))
	{
		fbo = (struct filebaseobject*)(m_pBaseArray->Data()+i);
		
		if ( fbo->m_nHashIndex == hbi )
		{
			*bi = i/sizeof(struct filebaseobject);
			return true;
		}
	}
	
	return false;
}

/** */
bool CSearchIndex::FindHashBaseIndex( struct filebaseobject * fbo, unsigned long * hbi )
{
	unsigned long i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+i);
		
		if ( Compare(fbo,hbo) )
		{
			*hbi = i;
			return true;
		}
	}
	
	return false;
}

/** */
bool CSearchIndex::FindHashBaseIndexDuringUpdate( struct filebaseobject * fbo, unsigned long * hbi )
{
	unsigned long i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pUpdatingHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+i);
		
		if ( CompareDuringUpdate(fbo,hbo) )
		{
			*hbi = i;
			return true;
		}
	}
	
	return false;
}

/** */
bool CSearchIndex::Compare( struct filebaseobject * fbo, struct hashbaseobject * hbo )
{
	if ( fbo->m_nSize != hbo->m_nSize )
		return false;
	if ( fbo->m_tModTime != hbo->m_tModTime )
		return false;

	CString s1,s2;
	
	s1 = (char*)m_pFileBaseArray->Data()+fbo->m_nFileIndex;
	s2 = (char*)m_pHashFileBaseArray->Data()+hbo->m_nFileIndex;
	
	if ( s1 != s2 )
		return false;
	
	s1 = (char*)m_pPathBaseArray->Data()+fbo->m_nPathIndex;
	s2 = (char*)m_pHashPathBaseArray->Data()+hbo->m_nPathIndex;
	
	if ( s1 != s2 )
		return false;

	return true;
}

/** */
bool CSearchIndex::CompareDuringUpdate( struct filebaseobject * fbo, struct hashbaseobject * hbo )
{
	if ( fbo->m_nSize != hbo->m_nSize )
		return false;
	if ( fbo->m_tModTime != hbo->m_tModTime )
		return false;

	CString s1,s2;
	
	s1 = (char*)m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex;
	s2 = (char*)m_pUpdatingHashFileBaseArray->Data()+hbo->m_nFileIndex;
	
	if ( s1 != s2 )
		return false;
	
	s1 = (char*)m_pUpdatingPathBaseArray->Data()+fbo->m_nPathIndex;
	s2 = (char*)m_pUpdatingHashPathBaseArray->Data()+hbo->m_nPathIndex;
	
	if ( s1 != s2 )
		return false;

	return true;
}

/** */
void CSearchIndex::AddHashIndex( unsigned long filebaseindex, unsigned char * hash, unsigned char * leaves, unsigned long lsize )
{
	struct filebaseobject fbo;
	struct hashbaseobject hbo;
	char * c;
	unsigned long hi,hbi;
	ulonglong storedlsize;
	
	if ( !GetFileBaseObjectDuringUpdate(filebaseindex,&fbo) )
	{
		return;
	}
	
	hi = 0;
	if ( FindHashDuringUpdate(hash,&hi) )
	{
		if ( HashBaseIndexFromHashIndexDuringUpdate(hi,&hbi) )
		{
			if ( CompareDuringUpdate(&fbo,(struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+hbi)) )
			{
				printf("hash found\n");
				fbo.m_nHashIndex = hbi;
				UpdateIndex(filebaseindex,&fbo);
				return;
			}
		}
	}
	
	hbo.m_nSize = fbo.m_nSize;
	hbo.m_tModTime = fbo.m_tModTime;
	hbo.m_nFileIndex = m_pUpdatingHashFileBaseArray->Size();
	hbo.m_nPathIndex = m_pUpdatingHashPathBaseArray->Size();
	hbo.m_nHashIndex = m_pUpdatingHashIndex->Size();
	
	if ( (lsize == 0) || (leaves == 0) )
	{
		hbo.m_nHashLeavesIndex = (unsigned long)-1;
	}
	else
	{
		if ( m_pNewHashLeaves == 0 )
		{
			m_pNewHashLeaves = new CByteArray();
			hashleavessize = CDir().getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",false);
		}
		
		storedlsize = lsize;
		hbo.m_nHashLeavesIndex = hashleavessize + m_pNewHashLeaves->Size();
		m_pNewHashLeaves->Append((unsigned char*)&storedlsize,sizeof(storedlsize));
		m_pNewHashLeaves->Append(leaves,lsize);
	}
	
	fbo.m_nHashIndex = m_pUpdatingHashBaseArray->Size();

	m_pUpdatingHashBaseArray->Append((const char*)&hbo, sizeof(struct hashbaseobject));
	c = (char*)m_pUpdatingFileBaseArray->Data()+fbo.m_nFileIndex;
	m_pUpdatingHashFileBaseArray->Append(c,strlen(c)+1);
	c = (char*)m_pUpdatingPathBaseArray->Data()+fbo.m_nPathIndex;
	m_pUpdatingHashPathBaseArray->Append(c,strlen(c)+1);
	m_pUpdatingHashIndex->Append( hash, dcpp::TigerTree::BYTES );

	UpdateIndex(filebaseindex,&fbo);
}

/** */
CString CSearchIndex::GetFileName( unsigned long i )
{
	CString s;
	struct filebaseobject * fbo;

	if ( (i*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
	{
		fbo = (struct filebaseobject *) (m_pBaseArray->Data()+(i*sizeof(struct filebaseobject)));

		s = (char*) (m_pFileBaseArray->Data()+fbo->m_nFileIndex);
	}

	return s;
}

/** */
CString CSearchIndex::GetFileNameDuringUpdate( unsigned long i )
{
	CString s;
	struct filebaseobject * fbo;

	if ( (i*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		fbo = (struct filebaseobject *) (m_pUpdatingBaseArray->Data()+(i*sizeof(struct filebaseobject)));

		s = (char*) (m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex);
	}

	return s;
}

/** */
CByteArray * CSearchIndex::GetHashLeaves( CString tth )
{
	CByteArray * ba = 0;
	unsigned long hi = 0, hbi = 0;
	
	/*
	 * this should not have been a ulonglong but we are stuck with it since sizeof(ulonglong) bytes
	 * are present before the hash leaf data
	 */
	ulonglong lsize = 0;
	
	unsigned long hli = 0;
	struct hashbaseobject * hbo = 0;
	CBase32 base32;
	CByteArray dst;
	
	if (base32.Decode(&dst, &tth) != dcpp::TigerTree::BYTES)
	{
		DPRINTF("GetHashLeaves: Decoded TTH length != dcpp::TigerTree::BYTES\n");
		return 0;
	}
	
	if ( FindHash(dst.Data(), &hi) )
	{
		//DPRINTF("GetHashLeaves: HashIndex=%llu\n", hi);
		
		if ( HashBaseIndexFromHashIndex( hi, &hbi ) )
		{
			//DPRINTF("GetHashLeaves: HashBaseIndex=%llu\n", hbi);
			hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+hbi);
			
			// hash leaves index
			hli = hbo->m_nHashLeavesIndex;
			
			if (hli == (unsigned long)-1)
			{
				DPRINTF("GetHashLeaves: No Leaves available.\n");
				return 0;
			}
			
			//DPRINTF("GetHashLeaves: HashLeavesIndex=%llu\n", hli);
			
			/* Old code for in memory hash leaves*/
			/* lsize = * (ulonglong *) (m_pHashLeaves->Data()+hli);
			
			//DPRINTF("GetHashLeaves: lsize=%llu\n", lsize);
			
			ba = new CByteArray();
			ba->Append( (m_pHashLeaves->Data()+(hli + sizeof(lsize))), lsize );
			
			return ba; */
			
			/* New code for on disk hash leaves */
			CFile leaffile;
			
			if ( leaffile.Open(CConfig::Instance()->GetConfigPath()+"hashleaves.bin", IO_RAW | IO_READONLY) == false )
			{
				printf("GetHashLeaves: cannot open hashleaves.bin\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				return 0;
			}
			
			if ( leaffile.Seek(hli, SEEK_SET) == false )
			{
				printf("GetHashLeaves: seek failed, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				return 0;
			}
			
			long read1 = 0, read2 = 0;
			read1 = leaffile.Read( (char *)&lsize, sizeof(lsize) );
			if (read1 != sizeof(lsize))
			{
				printf("GetHashLeaves: error didn't read leaves size, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				return 0;
			}
			
			//DPRINTF("GetHashLeaves: lsize=%llu\n", lsize);
			
			ba = new CByteArray( lsize );
			
			read2 = leaffile.Read( (char*)ba->Data(), lsize );
			if (read2 != lsize)
			{
				printf("GetHashLeaves: error didn't read enough data, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				delete ba;
				return 0;
			}
			
			leaffile.Close();
			
			return ba;
		}
		else
		{
			printf("HashBaseIndexFromHashIndex( hi, &hbi ) failed for %lu\n", hi);
			return 0;
		}
	}
	else
	{
		//printf("FindHash for %s failed\n", tth.Data());
		return 0;
	}
}

/** */
bool CSearchIndex::SaveNewHashLeaves()
{
	bool res;
	
	if (m_pNewHashLeaves != 0)
	{
		
		CFile file;
		res = file.Open( CConfig::Instance()->GetConfigPath()+"hashleaves.bin",
			   IO_RAW | IO_WRITEONLY | IO_APPEND | IO_CREAT,
			   MO_IRUSR | MO_IWUSR
		);
		
		if ( res == false )
		{
			printf("CSearchIndex::SaveNewHashLeaves: open hashleaves.bin failed\n");
		}
		else
		{
			long wrote = file.Write( (const char*) m_pNewHashLeaves->Data(), m_pNewHashLeaves->Size() );
			
			if ( wrote == m_pNewHashLeaves->Size() )
			{
				res = true;
			}
			else
			{
				printf("CSearchIndex::SaveNewHashLeaves: only wrote %ld out of %ld bytes\n",wrote,m_pNewHashLeaves->Size());
				if ( wrote > 0 )
				{
					printf("CSearchIndex::SaveNewHashLeaves: hashleaves.bin has been corrupted, do /rebuild\n");
				}
				res = false;
			}
			
			file.Close();
			
			/* If data was written to hashleaves.bin, even if only some, delete m_pNewHashLeaves */
			if ( wrote > 0 )
			{
				delete m_pNewHashLeaves;
				m_pNewHashLeaves = 0;
			}
		}
	}
	else
	{
		//DPRINTF("CSearchIndex::SaveNewHashLeaves: nothing to save\n");
		res = false;
	}
	
	return res;
}

/** */
void CSearchIndex::RebuildLists()
{
	CByteArray * pNewBaseArray     = new CByteArray();
	CByteArray * pNewFileBaseArray = new CByteArray();
	CByteArray * pNewPathBaseArray = new CByteArray();
	
	CByteArray * pNewHashBaseArray     = new CByteArray();
	CByteArray * pNewHashFileBaseArray = new CByteArray();
	CByteArray * pNewHashPathBaseArray = new CByteArray();
	CByteArray * pNewHashIndex         = new CByteArray();
	CByteArray * pNewHashLeaves        = 0;
	
	// we cannot do anything with the hash leaves if hashing is disabled...
	// and just in case someone changes the setting while this is running, store the value
	bool disablehashlist = CConfig::Instance()->GetDisableHashList();
	if ( disablehashlist == false )
	{
		pNewHashLeaves = new CByteArray();
	}
	
	unsigned long count = IndexCount();
	
	if ( dclibVerbose() > 0 )
	{
		printf("##### Before Rebuild #####\n");
		PrintDatabaseStats();
		printf("##########################\n");
	}
	
	CFile file;
	CDir cdir;
	CBase32 b32;
	
	for (unsigned long i = 0; i < count; i++)
	{
		if ( CFileManager::Instance()->Stopped() )
		{
			printf("RebuildLists stopped\n");
			
			delete pNewBaseArray;
			delete pNewFileBaseArray;
			delete pNewPathBaseArray;
			
			delete pNewHashBaseArray;
			delete pNewHashFileBaseArray;
			delete pNewHashPathBaseArray;
			delete pNewHashIndex;
			if ( pNewHashLeaves != 0 )
			{
				delete pNewHashLeaves;
			}
			
			return;
		}
		
		struct filebaseobject fbo;
		
		if ( GetFileBaseObject( i, &fbo ) == false )
		{
			DPRINTF("RebuildLists: failed to get filebaseobject for %lu\n", i);
			continue;
		}
		
		CString vpath = (char*) m_pPathBaseArray->Data()+fbo.m_nPathIndex;
		
		// top level folders have "/" as the path and their name as the filename
		CString filenamewithpath;
		
		if ( vpath.Length() > 1 )
		{
			filenamewithpath = vpath;
			filenamewithpath += DIRSEPARATOR;
		}
		
		CString filename = (char*) m_pFileBaseArray->Data()+fbo.m_nFileIndex;
		filenamewithpath += filename;
		
		CString realpath = CConfig::Instance()->AliasToPath(filenamewithpath);
		
		DPRINTF( "'%s' ", realpath.Data() );
		
		// check if file no longer shared
		if ( realpath.IsEmpty() )
		{
			DPRINTF( "%s no longer shared, entry removed\n", filenamewithpath.Data() );
			continue;
		}
		
		// if file exists
		if ( cdir.IsDir(realpath,false) )
		{
			DPRINTF("is existing folder\n");
			fbo.m_nFileIndex = pNewFileBaseArray->Size();
			fbo.m_nPathIndex = pNewPathBaseArray->Size();
			fbo.m_nHashIndex = (unsigned long)-1;
			
			pNewBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
			pNewFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
			pNewPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
		}
		else if ( file.Open(realpath, IO_RAW | IO_READONLY) )
		{
			file.Close();
			DPRINTF("exists ");
			
			fbo.m_nFileIndex = pNewFileBaseArray->Size();
			fbo.m_nPathIndex = pNewPathBaseArray->Size();
			
			// get TTH and hash leaves
			if ( fbo.m_nHashIndex != (unsigned long)-1 )
			{
				struct hashbaseobject * oldhbo = (struct hashbaseobject *) (m_pHashBaseArray->Data()+fbo.m_nHashIndex);

				CByteArray dst,src;
				src.Append( m_pHashIndex->Data()+oldhbo->m_nHashIndex, dcpp::TigerTree::BYTES );
				b32.Encode( &dst, &src );
				CString tth;
				tth.Set( (const char*)dst.Data(), dst.Size() );
				
				DPRINTF("tth=%s\n", tth.Data() );
				
				CByteArray * hashleaves = 0;
				ulonglong lsize = 0;
				
				if ( disablehashlist == false )
				{
					hashleaves = GetHashLeaves( tth );
					
					if ( hashleaves != 0 )
					{
						lsize = hashleaves->Size();
						
						if ( CFileHasher::ValidateHashLeaves( &src, hashleaves, cdir.getFileSize(realpath,false) ) == false )
						{
							delete hashleaves;
							hashleaves = 0;
							lsize = 0;
						}
					}
					
					if ( lsize == 0 )
					{
						printf("leaves not found or corrupt for %s, generating\n", tth.Data());
						CFileHasherThread hasher( realpath );
						hasher.Start();
						while ( hasher.GetStatus() == efhsWorking )
						{
							if ( CFileManager::Instance()->Stopped() )
							{
								printf("RebuildLists: CFileManager stopped, stopping hasher\n");
								hasher.StopHashing();
								
								delete pNewBaseArray;
								delete pNewFileBaseArray;
								delete pNewPathBaseArray;
								
								delete pNewHashBaseArray;
								delete pNewHashFileBaseArray;
								delete pNewHashPathBaseArray;
								delete pNewHashIndex;
								if ( pNewHashLeaves != 0 )
								{
									delete pNewHashLeaves;
								}
								
								return;
							}
#ifdef WIN32
							Sleep(100);
#else
							usleep(100000);
#endif
						}
				
						if ( hasher.GetStatus() == efhsFinished )
						{
							if ( tth != hasher.GetHashRoot() )
							{
								printf("RebuildLists: tth root mismatch, old=%s new=%s ,fixing\n", tth.Data(), hasher.GetHashRoot().Data());
								src.SetSize(0);
								src.Append( hasher.HashRootDirect()->Data(), hasher.HashRootDirect()->Size() );
							}
						
							if ( hashleaves == 0 )
							{
								hashleaves = new CByteArray();
							}
							
							hashleaves->SetSize(0);
							lsize = hasher.HashLeavesDirect()->Size();
							//printf("new lsize=%llu new hli=%lu (unsigned long)-1=%lu\n", lsize, pNewHashLeaves->Size(), (unsigned long)-1 );
							hashleaves->Append( hasher.HashLeavesDirect()->Data(), lsize );
						}
						else
						{
							printf("RebuildLists: problem generating hash data\n");
							lsize = 0;
							hashleaves->SetSize(0);
						}
					}
				}
				
				fbo.m_nHashIndex = pNewHashBaseArray->Size();
				
				struct hashbaseobject hbo;
				hbo.m_nSize    = fbo.m_nSize;
				hbo.m_tModTime = fbo.m_tModTime;

				hbo.m_nPathIndex = pNewHashPathBaseArray->Size();
				hbo.m_nFileIndex = pNewHashFileBaseArray->Size();
				hbo.m_nHashIndex = pNewHashIndex->Size();
				
				if ( disablehashlist == false )
				{
					if ( lsize > 0 )
					{
						hbo.m_nHashLeavesIndex = pNewHashLeaves->Size();
					}
					else
					{
						hbo.m_nHashLeavesIndex = (unsigned long)-1;
					}
					
				}
				else
				{
					hbo.m_nHashLeavesIndex = oldhbo->m_nHashLeavesIndex;
				}
				
				pNewHashBaseArray->Append( (const char*)&hbo, sizeof(struct hashbaseobject) );
				pNewHashPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
				pNewHashFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
				pNewHashIndex->Append( src.Data(), dcpp::TigerTree::BYTES );
				
				if ( lsize > 0 )
				{
					pNewHashLeaves->Append( (unsigned char*)&lsize, sizeof(lsize) );
					pNewHashLeaves->Append( hashleaves->Data(), lsize );
				}
				
				if ( hashleaves != 0 )
				{
					delete hashleaves;
				}
			}
			else
			{
				DPRINTF("no tth\n");
			}
			
			pNewBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
			pNewFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
			pNewPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
		}
		else
		{
			DPRINTF("does not exist, entry removed\n");
		}
	}
	
	// delete old lists
	delete m_pBaseArray;
	delete m_pFileBaseArray;
	delete m_pPathBaseArray;
	
	delete m_pHashBaseArray;
	delete m_pHashFileBaseArray;
	delete m_pHashPathBaseArray;
	delete m_pHashIndex;
	
	// replace old lists with new lists
	m_pBaseArray     = pNewBaseArray;
	m_pFileBaseArray = pNewFileBaseArray;
	m_pPathBaseArray = pNewPathBaseArray;
	
	m_pHashBaseArray     = pNewHashBaseArray;
	m_pHashFileBaseArray = pNewHashFileBaseArray;
	m_pHashPathBaseArray = pNewHashPathBaseArray;
	m_pHashIndex         = pNewHashIndex;
	
	if ( disablehashlist == false )
	{
		bool res = pNewHashLeaves->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashleaves.bin");
		if ( res == false )
		{
			/*
			 * most likely out of disk space, a permissions problem
			 * or filesystem remounted read only due to kernel detecting an error
			 */
			printf("Failed to save rebuilt hash leaf data, hash leaf data destroyed\n");
		}
		delete pNewHashLeaves;
	}
	
	if ( dclibVerbose() > 0 )
	{
		printf("##### After Rebuild #####\n");
		PrintDatabaseStats();
		printf("#########################\n");
	}
}

/** */
void CSearchIndex::PrintDatabaseStats()
{
	printf("BaseArray size=%lu itemsize=%d items=%lu\n", m_pBaseArray->Size(), sizeof(struct filebaseobject), m_pBaseArray->Size()/sizeof(struct filebaseobject) );
	printf("FileBaseArray size=%lu\n", m_pFileBaseArray->Size());
	printf("PathBaseArray size=%lu\n", m_pPathBaseArray->Size());
	
	printf("CaseFoldedBase size=%lu itemsize=%d items=%lu\n", m_pCaseFoldedBase->Size(), sizeof(unsigned long), m_pCaseFoldedBase->Size()/sizeof(unsigned long) );
	printf("CaseFoldedData size=%lu\n", m_pCaseFoldedData->Size());
	
	printf("HashBaseArray size=%lu itemsize=%d items=%lu\n", m_pHashBaseArray->Size(), sizeof(struct hashbaseobject), m_pHashBaseArray->Size()/sizeof(struct hashbaseobject));
	printf("HashFileBaseArray size=%lu\n", m_pHashFileBaseArray->Size());
	printf("HashPathBaseArray size=%lu\n", m_pHashPathBaseArray->Size());
	printf("HashIndex size=%lu itemsize=%d items=%lu\n", m_pHashIndex->Size(), dcpp::TigerTree::BYTES, m_pHashIndex->Size()/dcpp::TigerTree::BYTES);
	
	printf("HashLeaves size=%llu (on disk)\n", CDir().getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",false));
}
